﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DateDiffInDays.DateDiff
{
    /// <summary>
    /// Date difference calculation by getting the delay in days of both dates to an origin, 
    /// then get the difference of both delays.
    /// Origin is date: 01-01-01
    /// </summary>
    class DateDiffByDiffToOrigin : IDateDiff
    {
        /// <summary>
        /// each entry corresponds to the number of days in a non-leap year till the month = index. Zero-index not used.
        /// </summary>
        private static readonly int[] daysUpToMonth = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
        /// <summary>
        /// each entry corresponds to the number of days in a leap year till the month = index. Zero-index not used.
        /// </summary>
        private static readonly int[] daysUpToMonthLeapYear = { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335 };

        public DateDiffByDiffToOrigin(int y1, int m1, int d1, int y2, int m2, int d2) :
            base(y1, m1, d1, y2, m2, d2)
        {
            /* empty, see base constructor */
        }

        public override int GetDifference()
        {
            Logger.GetLogger().Log("Calling GetDaysOffsetFromOrigin with " + yearPreceding + "-" +
                monthPreceding + "-" + dayPreceding);
            int daysOffset = GetDaysOffsetFromOrigin(yearPreceding, monthPreceding, dayPreceding);

            Logger.GetLogger().Log("Calling GetDaysOffsetFromOrigin with " + yearSubsequent + "-" +
                monthSubsequent + "-" + daySubsequent);
            int daysOffset2 = GetDaysOffsetFromOrigin(yearSubsequent, monthSubsequent, daySubsequent);

            Logger.GetLogger().Log("-----------------------");
            Logger.GetLogger().Log("daysOffset = " + daysOffset);
            Logger.GetLogger().Log("daysOffset2 = " + daysOffset2);

            Logger.GetLogger().Log("Result = (daysOffset2 - daysOffset) = " + (daysOffset2 - daysOffset));

            return (daysOffset2 - daysOffset) * (IsDate1AfterDate2 ? -1 : 1);
        }

        /// <summary>
        /// Get number of days from origin 01-01-01 to year-month-day.
        /// </summary>
        /// <param name="year"></param>
        /// <param name="month"></param>
        /// <param name="day"></param>
        /// <returns></returns>
        public static int GetDaysOffsetFromOrigin(int year, int month, int day)
        {
            if (DateDiffUtil.IsBeforeThan(year, month, day, 1, 1, 1))
            {
                throw new ArgumentException("Invalid date: " + year + "-" + month + "-" + day);
            }

            if (DateDiffUtil.IsLeapYear(year))
            {
                year--;
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: Year is leap year");
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: year = " + year);
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: numOfLeapsYear = " + 
                    (year / 4 - year / 100 + year / 400));

                int numOfLeapsYear = year / 4 - year / 100 + year / 400;

                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: year * 365 + numOfLeapsYear = " +
                    (year * 365 + numOfLeapsYear));
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: daysUpToMonthLeapYear[month - 1] + day - 1 = " +
                    (daysUpToMonthLeapYear[month - 1] + day - 1));
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: Result = " + 
                    (year * 365 + numOfLeapsYear + daysUpToMonthLeapYear[month - 1] + day - 1));

                return year * 365 + numOfLeapsYear + daysUpToMonthLeapYear[month - 1] + day - 1;
            }
            else
            {
                year--;

                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: Year is not a leap year");
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: year = " + year);
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: numOfLeapsYear = " + 
                    (year / 4 - year / 100 + year / 400));

                int numOfLeapsYear = year / 4 - year / 100 + year / 400;

                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: year * 365 + numOfLeapsYear = " +
                    (year * 365 + numOfLeapsYear));
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: daysUpToMonth[month - 1] + day - 1 = " +
                    (daysUpToMonth[month - 1] + day - 1));
                Logger.GetLogger().Log("GetDaysOffsetFromOrigin: Result = " + 
                    (year * 365 + numOfLeapsYear + daysUpToMonth[month - 1] + day - 1));

                return year * 365 + numOfLeapsYear + daysUpToMonth[month - 1] + day - 1;
            }
        }
    }
}
