﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DateDiffInDays.DateDiff
{
    /// <summary>
    /// Date difference calculation in three steps:
    /// - Calculate the remaing days in the preceding year.
    /// - Count the years between both dates and get the number of days.
    /// - Calculate the days in the last year until subsequent days.
    /// </summary>
    class DateDiffDirectThreeSteps : IDateDiff
    {
        private static readonly int[] daysUpToMonth = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
        private static readonly int[] daysUpToMonthLeapYear = { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335 };

        public DateDiffDirectThreeSteps(int y1, int m1, int d1, int y2, int m2, int d2) :
            base(y1, m1, d1, y2, m2, d2)
        {
            /* empty, see base constructor */
        }

        public override int GetDifference()
        {
            if (yearPreceding == yearSubsequent)
            {   /* same year, easy case */
                Logger.GetLogger().Log("Both years are equal");
                return GetDiffInDaysInSameYear(monthPreceding, dayPreceding, monthSubsequent, daySubsequent, DateDiffUtil.IsLeapYear(yearPreceding));
            }
            else
            {
                int resultDays = 0;

                /* handle first year which might be not a whole year */
                resultDays = GetRemainingDaysInYear(monthPreceding, dayPreceding, DateDiffUtil.IsLeapYear(yearPreceding));
                /* first year is now handled */
                yearPreceding++;
                Logger.GetLogger().Log("Step 1 : resultDays = " + resultDays);

                while (yearPreceding < yearSubsequent)
                {
                    Logger.GetLogger().Log(yearPreceding + " < " + yearSubsequent + " - add " +
                        (DateDiffUtil.IsLeapYear(yearPreceding) ? 366 : 365) + " years");
                    resultDays += DateDiffUtil.IsLeapYear(yearPreceding) ? 366 : 365;
                    yearPreceding++;
                }
                Logger.GetLogger().Log("Step 2 : resultDays = " + resultDays);

                resultDays += GetDaysInYearTillDate(monthSubsequent, daySubsequent, DateDiffUtil.IsLeapYear(yearSubsequent));
                Logger.GetLogger().Log("Step 3 : resultDays = " + resultDays);
                return resultDays * (IsDate1AfterDate2 ? -1 : 1);
            }
        }

        /// <summary>
        /// Get the difference in days between both dates, specified by day and month.
        /// Precondition: It is assumed that both dates lie in the same year.
        /// If date 2 lies before date 1, the result is negative.
        /// </summary>
        /// <param name="month1"></param>
        /// <param name="day1"></param>
        /// <param name="month2"></param>
        /// <param name="day2"></param>
        /// <returns></returns>
        public int GetDiffInDaysInSameYear(int month1, int day1, int month2, int day2, bool isLeapYear)
        {
            if (isLeapYear)
            {
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: Year is leap year");
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: daysUpToMonthLeapYear[month2 - 1] = " + daysUpToMonthLeapYear[month2 - 1]);
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: daysUpToMonthLeapYear[month1 - 1] = " + daysUpToMonthLeapYear[month1 - 1]);
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: (day2 - day1) = " + (day2 - day1));
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: Result = " + (daysUpToMonthLeapYear[month2 - 1] - daysUpToMonthLeapYear[month1 - 1] +
                      (day2 - day1)));

                return daysUpToMonthLeapYear[month2 - 1] - daysUpToMonthLeapYear[month1 - 1] +
                      (day2 - day1);
            }
            else
            {
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: Year is not a leap year");
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: daysUpToMonth[month2 - 1] = " + daysUpToMonth[month2 - 1]);
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: daysUpToMonth[month1 - 1] = " + daysUpToMonth[month1 - 1]);
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: (day2 - day1) = " + (day2 - day1));
                Logger.GetLogger().Log("GetDiffInDaysInSameYear: Result = " + (daysUpToMonth[month2 - 1] - daysUpToMonth[month1 - 1] +
                      (day2 - day1)));

                return daysUpToMonth[month2 - 1] - daysUpToMonth[month1 - 1] +
                      (day2 - day1);
            }
        }

        /// <summary>
        /// Get the number of remaining days in given year, that is from month-day until 12-31.
        /// </summary>
        /// <param name="month"></param>
        /// <param name="day"></param>
        /// <param name="isLeapYear"></param>
        /// <returns></returns>
        public int GetRemainingDaysInYear(int month, int day, bool isLeapYear)
        {
            if (isLeapYear)
            {
                Logger.GetLogger().Log("GetRemainingDaysInYear: Year is leap year");
                Logger.GetLogger().Log("GetRemainingDaysInYear: daysUpToMonthLeapYear[month - 1] + day = " + 
                    (daysUpToMonthLeapYear[month - 1] + day));
                Logger.GetLogger().Log("GetRemainingDaysInYear: Result = " + 
                    (366 - (daysUpToMonthLeapYear[month - 1] + day)));

                return 366 - (daysUpToMonthLeapYear[month - 1] + day);
            }
            else
            {
                Logger.GetLogger().Log("GetRemainingDaysInYear: Year is not a leap year");
                Logger.GetLogger().Log("GetRemainingDaysInYear: daysUpToMonth[month - 1] + day = " + 
                    (daysUpToMonth[month - 1] + day));
                Logger.GetLogger().Log("GetRemainingDaysInYear: Result = " + 
                    (365 - (daysUpToMonth[month - 1] + day)));

                return 365 - (daysUpToMonth[month - 1] + day);
            }
        }

        /// <summary>
        /// Get the number of days of given date, that is from 01-01 to month-day.
        /// </summary>
        /// <param name="month"></param>
        /// <param name="day"></param>
        /// <param name="isLeapYear"></param>
        /// <returns></returns>
        public int GetDaysInYearTillDate(int month, int day, bool isLeapYear)
        {
            if (isLeapYear)
            {
                Logger.GetLogger().Log("GetDaysInYearTillDate: Year is leap year");
                Logger.GetLogger().Log("GetDaysInYearTillDate: Result = " +
                    (daysUpToMonthLeapYear[month - 1] + day));

                return daysUpToMonthLeapYear[month - 1] + day;
            }
            else
            {
                Logger.GetLogger().Log("GetDaysInYearTillDate: Year is not a leap year");
                Logger.GetLogger().Log("GetDaysInYearTillDate: Result = " +
                    (daysUpToMonth[month - 1] + day));
                return daysUpToMonth[month - 1] + day;
            }
        }
    }
}
